
(*<$Description$> Preventa Support library *
 * This FUNCTION block reads a binary sequence sent over the Preventa Safety Module auxiliary output Z1. 
 * This sequence consists OF two parts:
 * a start sequence OF 4 BITS (800ms) AND a diagnostic code seqeunce OF 6 BITS (1200ms).
 * Each bit duration being 200ms AND the whole sequence duration 2sec. The Z1 signal needs TO be monitored 
 * FOR a signal change every 200ms. The tasl yclte TO accomplish this needs t be at least 50ms OR less. 
 * The FUNCTION block needs TO verify this condition BY the input I_timTaskCycle. 
 * Each diagnostic code has TO be detected two times in a row (confirmed) before the corresponding BIT in Q_stDiagCode is set.
 * ---------------------------------------------------------
 * Called functions/FUNCTION blocks:
 * - FB_CheckCallInInterval
 * - TON
 * - R_TRIG </$Description$>*)
FUNCTION_BLOCK FB_PreventaDiag
    (* Reset temp values *)
    xErrorCode := FALSE;
    (* Enable/ disable function block *)
    IF NOT I_xEnable THEN	
        // Start values
        iState 			:= C_iStateIdle;
        iPreviousBit	:= -1;
        iActualBit 		:= 0;
        xNewStart       := TRUE;
        xFrameBuffer[1] := FALSE;
        xFrameBuffer[2] := FALSE;
        xFrameBuffer[3] := FALSE;
        xFrameBuffer[4] := FALSE;
        xFrameBuffer[5] := FALSE;
        xFrameBuffer[6] := FALSE;
        xFrameBuffer[7] := FALSE;
        xFrameBuffer[8] := FALSE;
        xFrameBuffer[9] := FALSE;
        xFrameBuffer[10] := FALSE;
        TON_TickerS(IN := FALSE);
        TON_TimeoutS(IN := FALSE);
        R_Trigger(CLK := I_xDiagSignal);
        xEdgeDetected := FALSE;
        xStartSeqFound  := FALSE;
        
        xShortCircuitOn     := FALSE;
        xPowerOff           := FALSE;
        xSendValidOnce      := FALSE;
    ELSE
	
        (* Edge detection to synchronize with the signal and check for timeout *)
        R_Trigger(CLK:=I_xDiagSignal);
        xEdgeDetected := R_Trigger.Q;

        (* Timer for timeout if no more edges in the signal detected -> Power Off / Wire Open *)
        IF (iState = C_iStateSearch) OR (iState = C_iStateCode) THEN
            TON_TimeoutS( IN:=NOT xEdgeDetected,PT:=C_timTimeOutStart);
            xTon_TimeoutS_Q := TON_TimeoutS.Q; (* Output parameter 'ET' not neccessary to read out *)
        ELSE
            TON_TimeoutS(IN := FALSE);	
            xTon_TimeoutS_Q	:= FALSE;
        END_IF

        // Timer to schedule the capture action on the signal
        IF (iState = C_iStateSearch) OR (iState = C_iStateCode) THEN
            (* Time ticker for binary sequence *)
            TON_TickerS(IN := TRUE,
                PT     := C_timSeqTotalMax); 
            xTON_TickerS_Q     := TON_TickerS.Q; 
            timTON_TickerS_ET  := TON_TickerS.ET;	
        ELSE
            TON_TickerS(IN  := FALSE);
        END_IF
	
        (* Allign the capture actions in the middle of each bit *)
        IF xEdgeDetected THEN
            iEdgeOffset := 100 - ((TIME_TO_INT(timTON_TickerS_ET)) MOD 200);
            IF iEdgeOffset < 0 THEN
                iEdgeOffset := iEdgeOffset + 200;
            END_IF
            // edge detected -> no power fail or shortcut
            xShortCircuitOn := FALSE;
            xPowerOff := FALSE;
            xSendValidOnce := FALSE;
            IF xNewStart THEN
                xNewStart := FALSE;
                TON_TickerS(IN := FALSE);
                TON_TimeoutS(IN := FALSE);
                timTON_TickerS_ET := T#0S;
                iEdgeOffset := 100;
                iEdgeOffset2 := 100;
                xFrameBuffer[1] := FALSE;
                xFrameBuffer[2] := FALSE;
                xFrameBuffer[3] := FALSE;
                xFrameBuffer[4] := FALSE;
                xFrameBuffer[5] := FALSE;
                xFrameBuffer[6] := FALSE;
                xFrameBuffer[7] := FALSE;
                xFrameBuffer[8] := FALSE;
                xFrameBuffer[9] := FALSE;
                xFrameBuffer[10] := FALSE;
            END_IF
        END_IF
	
        iEdgeOffset2 := (iEdgeOffset + iEdgeOffset2 * 4) / 5;
        
        // set the capture signal if a new bit has to be captured
        xCaptureSignal := FALSE;
        xNewFrame := FALSE;
        iActualBit := (TIME_TO_INT(timTON_TickerS_ET) + iEdgeOffset2) / C_iBitDuration;
        IF iActualBit <> iPreviousBit THEN
            IF iActualBit > 0 THEN
                xCaptureSignal := TRUE;
                IF NOT xFrameBuffer[7] AND NOT xFrameBuffer[8] AND xFrameBuffer[9] AND NOT xFrameBuffer[10] AND NOT xStartSeqFound THEN
                    xStartSeqFound := TRUE;
                    TON_TickerS(IN := FALSE);
                END_IF
                IF iActualBit = 10 AND xStartSeqFound THEN
                    xNewFrame := TRUE;
                    xLockSearch := FALSE;
                    wFrame := 0;
                    wFrame.0 := xFrameBuffer[1];
                    wFrame.1 := xFrameBuffer[2];
                    wFrame.2 := xFrameBuffer[3];
                    wFrame.3 := xFrameBuffer[4];
                    wFrame.4 := xFrameBuffer[5];
                    wFrame.5 := xFrameBuffer[6];
                    wFrame.6 := xFrameBuffer[7];
                    wFrame.7 := xFrameBuffer[8];
                    wFrame.8 := xFrameBuffer[9];
                    wFrame.9 := xFrameBuffer[10];
                    TON_TickerS(IN:= FALSE);
                END_IF
            END_IF
            iPreviousBit := iActualBit;
        END_IF

        IF xCaptureSignal THEN
            xFrameBuffer[1] := xFrameBuffer[2];
            xFrameBuffer[2] := xFrameBuffer[3];
            xFrameBuffer[3] := xFrameBuffer[4];
            xFrameBuffer[4] := xFrameBuffer[5];
            xFrameBuffer[5] := xFrameBuffer[6];
            xFrameBuffer[6] := xFrameBuffer[7];
            xFrameBuffer[7] := xFrameBuffer[8];
            xFrameBuffer[8] := xFrameBuffer[9];
            xFrameBuffer[9] := xFrameBuffer[10];
            xFrameBuffer[10] := I_xDiagSignal;
        END_IF
	
        // Timeout, no more edges within the signal
        xSendValidOnce := FALSE;
        IF xTon_TimeoutS_Q THEN
            xLockSearch := TRUE;
            IF iState <> 0 THEN
                xSendValidOnce := TRUE;
            END_IF
            iState := C_iStateIdle;
            dwCode1_32 := 0;
            dwCode33_64 := 0;
            IF I_xDiagSignal THEN
                // signal is high -> Shortcut
                xShortCircuitOn := TRUE;
                xPowerOff := FALSE;
                xSendValidOnce := FALSE;
            ELSE
                // signal is low -> Wire Open or Power Off
                xShortCircuitOn := FALSE;
                xPowerOff := TRUE;
                xInternValid   := FALSE;
            END_IF
        END_IF
	
        (* wait for timer to reach the end of the actual frame, then restart *)
        IF TON_TickerS.Q THEN
            //iEdgeOffset := 0;
            //iEdgeOffset2 := 0;
            (* Reset for a new start *)	
            TON_TickerS(IN:= FALSE);
        END_IF

    END_IF

    (*  *)

    (* Check FB call in each cycle.
    * Internal a timer runs with the defined TaskCycle AND check with a tolerance IF time
    * elapsed in next cycle *)
    fbCheckCall(I_timCycleTime  := I_timTaskCycle,
        I_xEnable       := NOT (C_iStateIdle = iState),
        );
    xErrorFbCall := fbCheckCall.Q_xCallFailed; 

    (* If one or more call missed, a error will be set, until the call is detected in the correct cycle *)
    IF xErrorFbCall THEN
        iState   := C_iStateError;
        wErrorId := C_wERROR_MissedCall;
    END_IF 

    (* State machine *)
    CASE iState OF
        C_iStateIdle: (* If function block is not enabled, state IDLE is active *)	
		
            // Default output values 		
            xInternBusy 		:= FALSE;
            xInternValid 		:= FALSE;
            xInternError		:= FALSE; // Reset internal FB status variables, if FB is disabled
            dwCode1_32      := 16#0;
            dwCode33_64     := 16#0;
            byCodeSeqOut    := 16#0;
            wErrorId	    := C_wERROR_none;
            xErrorCode		:= FALSE;
            TON_TimeoutS(IN := FALSE);
            TON_TickerS(IN  := FALSE);
	
            IF I_xEnable THEN
                iState   := C_iStateInit;
            END_IF
		
        C_iStateInit : (* Initializing and check of start parameter *)
	
            (* Reset variables *)
            iActInvalidFrames := 0;
            xInternBusy := FALSE;
            IF NOT xSendValidOnce THEN
                xInternValid := FALSE;
            END_IF
            xLockSearch := TRUE;
            (* Reset local BIT arrays *)
            byCodeSequence 	:= 16#0;

            (* Parameter check: Read out cycle time to compare with necessary value *)	
            IF (I_timTaskCycle > C_timTaskCycleTime) THEN
                iState 		:= C_iStateError;
                wErrorId	:= C_wERROR_CycleTime;		
            ELSE
                timCycleTimeLastInit 	:= I_timTaskCycle;
                iState 					:= C_iStateSearch;
                (* Set state machine status *)
                xInternBusy     	:= TRUE;
                xInternError 		:= FALSE;
                xInternValid 		:= FALSE;
            END_IF

        C_iStateSearch : (* Searching the start of the frame. Wait until the first valid frame is inside the buffer *)
	
            xInternBusy := TRUE;
            xInternValid := FALSE;

            IF xShortCircuitOn OR xPowerOff THEN
                iActInvalidFrames := 0;
            ELSE
                IF xNewFrame THEN
                    iActInvalidFrames := iActInvalidFrames + 1;
                END_IF
            END_IF
	
            IF iActInvalidFrames > C_iMaxInvSearchFrames THEN
                iState 		:= C_iStateError;
                wErrorId 	:= C_wERROR_SearchTOut;
            END_IF 
	
            // Wait for a new bit captured
            IF xCaptureSignal THEN
                xCaptureSignal := FALSE;
                // check if the start sequence is valid
                IF NOT xFrameBuffer[1] AND NOT xFrameBuffer[2] AND xFrameBuffer[3] AND NOT xFrameBuffer[4] THEN
                    // check if the code is existing
                    byCodeSequence := 0;
                    byCodeSequence.5 := xFrameBuffer[5];
                    byCodeSequence.4 := xFrameBuffer[6];
                    byCodeSequence.3 := xFrameBuffer[7];
                    byCodeSequence.2 := xFrameBuffer[8];
                    byCodeSequence.1 := xFrameBuffer[9];
                    byCodeSequence.0 := xFrameBuffer[10];
                    iCode := BYTE_TO_INT(byCodeSequence);
			
                    fbCheckCode(I_iCode := iCode);
                    xCodeValid := fbCheckCode.Q_xValid;
                    dwCode1_32_tmp := fbCheckCode.Q_dwCode1_32;
                    dwCode33_64_tmp := fbCheckCode.Q_dwCode33_64;

                    IF xCodeValid AND NOT xLockSearch THEN
                        iState := C_iStateCode;
                        byCodeSeqOut := byCodeSequence;
                        dwCode1_32 := dwCode1_32_tmp;
                        dwCode33_64 := dwCode33_64_tmp;
                        byCode_Buffered := byCodeSequence;
                        dwCode1_32_Buffered := dwCode1_32_tmp;
                        dwCode33_64_Buffered := dwCode33_64_tmp;
                        TON_TickerS(IN:= FALSE);
                        xInternBusy := FALSE;
                        xInternValid := TRUE;
                        iActInvalidFrames := 0;
                    END_IF
                END_IF
            END_IF
	
        C_iStateCode : (* Wait until a new code is received *)
		
            xInternBusy := TRUE;
            xInternValid := FALSE;

            IF xNewFrame THEN
                xCodeValid := FALSE;
                // check if the start sequence is valid
                IF NOT xFrameBuffer[1] AND NOT xFrameBuffer[2] AND xFrameBuffer[3] AND NOT xFrameBuffer[4] THEN
                    // check if the code is existing
                    byCodeSequence := 0;
                    byCodeSequence.5 := xFrameBuffer[5];
                    byCodeSequence.4 := xFrameBuffer[6];
                    byCodeSequence.3 := xFrameBuffer[7];
                    byCodeSequence.2 := xFrameBuffer[8];
                    byCodeSequence.1 := xFrameBuffer[9];
                    byCodeSequence.0 := xFrameBuffer[10];
                    iCode := BYTE_TO_INT(byCodeSequence);
                    fbCheckCode(I_iCode := iCode);                    
                    xCodeValid := fbCheckCode.Q_xValid;
                    dwCode1_32_tmp := fbCheckCode.Q_dwCode1_32;
                    dwCode33_64_tmp := fbCheckCode.Q_dwCode33_64;
                END_IF
		
                IF xCodeValid THEN
                    // Code was valid, send to output
                    IF C_xDelayOneFrame THEN
                        // buffer this code, send the last one
                        byCodeSeqOut := byCode_Buffered;
                        dwCode1_32 := dwCode1_32_Buffered;
                        dwCode33_64 := dwCode33_64_Buffered;
                        byCode_Buffered := byCodeSequence;
                        dwCode1_32_Buffered := dwCode1_32_tmp;
                        dwCode33_64_Buffered := dwCode33_64_tmp;
                    ELSE
                        // send directly
                        byCodeSeqOut := byCodeSequence;
                        dwCode1_32 := dwCode1_32_tmp;
                        dwCode33_64 := dwCode33_64_tmp;
                    END_IF
			
                    xInternBusy := FALSE;
                    xInternValid := TRUE;
                    // decrement the counter of invalid frames, if above 0
                    IF iActInvalidFrames > 0 THEN
                        iActInvalidFrames := iActInvalidFrames - 1;
                    END_IF
                ELSE
                    // frame not valid. check if frame consists of 0 only -> power off
                    IF wFrame <> 0 THEN
                        // frame contains not only 0, frame is invalid! increment counter
                        IF iActInvalidFrames > C_iMaxInvCodeFrames THEN
                            // more than C_iMaxInvCodeFrames invalid frames received, error
                            iState := C_iStateError;
                            xInternError  := TRUE;
                            wErrorId 	  := C_wERROR_InvalidCode;	
                        ELSE
                            // increment counter for invalid frames
                            iActInvalidFrames := iActInvalidFrames + 1;
                        END_IF
                    ELSE
                        // only 0 -> Power Off, go back to search
                        xLockSearch := TRUE;
                        iState := C_iStateSearch;
                    END_IF
                END_IF
            END_IF

        C_iStateError : (* Error occured, set output and change back to state Init *)
            xInternError 	:= TRUE;
            xInternValid  	:= FALSE;
            iState 	    	:= C_iStateInit;
            // SET DEFAULT VALUES IF AN ERROR WAS DETECTED
            dwCode1_32 		:= 16#0;
            dwCode33_64 	:= 16#0;
            byCodeSequence 	:= 16#0;
            byCodeSeqOut    := 16#0;
		
            (* Error in value for state machine *)
        ELSE
            xInternError := TRUE;
            wErrorId     := C_wERROR_InvalidState;
    END_CASE

    (* Define outputs *)
    //Q_dwStateStatus := INT_TO_DWORD(iState);
    Q_xBusy  := xInternBusy;
    Q_xValid  := xInternValid OR xSendValidOnce;
    Q_xError := xInternError;
    
    (*Hide error id in case boolean error output is false*)
    IF NOT xInternError THEN
        Q_wErrorId := 0;
    ELSE
        Q_wErrorId        := wErrorId;
    END_IF

    //Diagnostic code data output structure
    Q_stDiagCode.xOperational 		:= dwCode33_64.9;
    Q_stDiagCode.xWaitMonStartFTrig := dwCode33_64.8;
    Q_stDiagCode.xWaitMonStartRTrig := dwCode33_64.7;
    Q_stDiagCode.xWaitStartUpTest	:= dwCode33_64.6;
    Q_stDiagCode.xInp63ChangeState	:= dwCode33_64.5;
    Q_stDiagCode.xInp62ChangeState	:= dwCode33_64.4;
    Q_stDiagCode.xInp53ChangeState	:= dwCode33_64.3;
    Q_stDiagCode.xInp52ChangeState	:= dwCode33_64.2;
    Q_stDiagCode.xInp43ChangeState	:= dwCode33_64.1;
    Q_stDiagCode.xInp42ChangeState	:= dwCode33_64.0;
    Q_stDiagCode.xInp33ChangeState	:= dwCode1_32.31;
    Q_stDiagCode.xInp32ChangeState	:= dwCode1_32.30;
    Q_stDiagCode.xInp23ChangeState	:= dwCode1_32.29;
    Q_stDiagCode.xInp22ChangeState	:= dwCode1_32.28;
    Q_stDiagCode.xInp13ChangeState	:= dwCode1_32.27;
    Q_stDiagCode.xInp12ChangeState	:= dwCode1_32.26;
    Q_stDiagCode.xIoffAllNOCoff		:= dwCode1_32.25;
    Q_stDiagCode.xIoffNOCoffDelayC  := dwCode1_32.24;
    Q_stDiagCode.xSyncFltTestInRun	:= dwCode1_32.23;
    Q_stDiagCode.xSyncFltWaitTest	:= dwCode1_32.22;
    Q_stDiagCode.xErrorS2x			:= dwCode1_32.21;
    Q_stDiagCode.xErrorS1x			:= dwCode1_32.20;
    Q_stDiagCode.xExtFltDelayCirc	:= dwCode1_32.19;
    Q_stDiagCode.xExtFltStartMonFdb := dwCode1_32.18;
    Q_stDiagCode.xExtFltIn63Circ	:= dwCode1_32.17;
    Q_stDiagCode.xExtFltIn62Circ	:= dwCode1_32.16;
    Q_stDiagCode.xExtFltIn53Circ	:= dwCode1_32.15;
    Q_stDiagCode.xExtFltIn52Circ	:= dwCode1_32.14;
    Q_stDiagCode.xExtFltIn43Circ	:= dwCode1_32.13;
    Q_stDiagCode.xExtFltIn42Circ	:= dwCode1_32.12;
    Q_stDiagCode.xExtFltIn33Circ	:= dwCode1_32.11;
    Q_stDiagCode.xExtFltIn32Circ	:= dwCode1_32.10;
    Q_stDiagCode.xExtFltIn23Circ	:= dwCode1_32.9;
    Q_stDiagCode.xExtFltIn22Circ	:= dwCode1_32.8;
    Q_stDiagCode.xExtFltIn13Circ	:= dwCode1_32.7;
    Q_stDiagCode.xExtFltIn12Circ	:= dwCode1_32.6;
    Q_stDiagCode.xErrorConfig 	    := dwCode1_32.5;
    Q_stDiagCode.xIntFltInExModule  := dwCode1_32.4;
    Q_stDiagCode.xIntFlt			:= dwCode1_32.3;
    Q_stDiagCode.xSupplyOutOfTol 	:= dwCode1_32.2;
    Q_xComWireInShort := xShortCircuitOn;
    Q_xComWireOpen := xPowerOff;
    Q_dwStatus := BYTE_TO_DWORD(byCodeSeqOut);
END_FUNCTION_BLOCK
