
(*<$Description$>Preventa Support library *
 * This FUNCTION block monitores the diagnostic results FOR maintenance information. 
 * The current diagnostic code status must be read in (analyzed BY the FB_PreventaDiag).
 * Further it monitors proof test executions FOR safety reasons.
 * ------------------------------------------------------
 * Called functions/FUNCTION blocks:
 * - R_TRIG: rising trigger OF a variable  </$Description$>*)
FUNCTION_BLOCK FB_PreventaMain
    (* Reset Temp Values *)
    xErrorResult1   := FALSE;
    xErrorResult2   := FALSE;	
    xTempError      := FALSE;
    xCountAllInputs := FALSE;
    xLastCodeWas0   := FALSE;
    (* Enable/ disable function block *) 
    IF NOT I_xEnable THEN
        (* Set help variables to default values - outputs to default values *)
        xTempError 		:= FALSE;
        wTempErrorId 	:= 16#0;	
        udiStatus_valid	:= DWORD_TO_UDINT(I_dwStatus);
        udiStatus_old	:= udiStatus_valid;
    ELSE		
        (* Define start values for loop counters for each new cycle *)
        diLoopIndex		:= 0;
        diLoopCountOut	:= 1;
        iLoopReset 		:= 1; 
        diLoopTrig 		:= 1;
        xOpExceeded		:= FALSE;
        wTempExceededId	:= 0;
        xRTrig_Operational_Q := FALSE;
        xTrig_CodeChanged := FALSE;
	
        IF I_xValid THEN
            udiStatus_valid := DWORD_TO_UDINT(I_dwStatus);
        END_IF
        (* read current time if not provided *)
        IF (DT_TO_UDINT(I_dtTestIntervalRef) = 0) THEN
            DTGetTime_0.enable := TRUE;
            DTGetTime_0();
            dtTestIntervalRef := DTGetTime_0.DT1;
        ELSE
            dtTestIntervalRef := I_dtTestIntervalRef;
        END_IF
    END_IF

    (* Start executing the maintenance tasks after FB was enabled *)
    IF I_xEnable THEN
	
        (* Check if the status of the module changed *)
        IF udiStatus_valid <> udiStatus_old THEN
            (* check if the status code changed to CMD41 (operational) *)
            IF udiStatus_valid = C_udiCMD_Operational THEN
                xRTrig_Operational_Q := TRUE;
            ELSE
                (* Check if one of the inputs is triggered by the actual CMD *)
                xTrig_CodeChanged := TRUE;
                (* Change from operational to All Inputs Off? -> count all inputs *)
                IF udiStatus_old = C_udiCMD_Operational AND udiStatus_valid = C_udiCMD_AllInpOff THEN
                    xCountAllInputs := TRUE;
                END_IF
            END_IF
            IF udiStatus_old = 0 THEN
                xLastCodeWas0 := TRUE;
            ELSE
                xLastCodeWas0 := FALSE;
            END_IF
            udiStatus_old := udiStatus_valid;
        END_IF

(* Exchanging the module is confirmed by an active input I_xNewModule Active. 
 A rising EDGE causes the reset OF module counters - Start timestamp, Array OF 
  proof tests AND the summarized operation counter *)
        RTrig_ModuleExch(CLK:= I_xNewModuleActive);
        xRTrig_ModuleExch_Q := RTrig_ModuleExch.Q;
					 
        IF xRTrig_ModuleExch_Q THEN
            (* Reset summarized operation counter *)
            R_udiCountSystem := 0; 
            (* Reset array of proof tests *)
            iLoopReset := 1; 
            FOR iLoopReset := 1 TO C_iProofTestStorage DO
                R_adtProof[iLoopReset] := DT#1970-1-1-0:0:0;
            END_FOR
            (* Set a new Start timestamp for activating the new module *)
            R_datTestStart  := DT_TO_DATE(dtTestIntervalRef ); // Date of the first comissioning of the module, which stay constant 
            R_adtProof[1]   := dtTestIntervalRef; // take over values of the very first test timestamp to the remanent array	
            (* comment: the counters of inputs/outptus and for the proccess will be reset by the single correspondign data structure reset input *)		
        END_IF

        (* Input parameter check *)
        diInpArrLimitMin := 1; (* Minimum value of input contact array *)
        diInpArrLimitMax := C_iNumOfInputCont; (*Maximum value of input contact array *)
        diOutArrLimitMin := 1; (* Minimum value of output contact array *)
        diOutArrLimitMax := C_iNumOfOutputCont; (* Maximum value of output contact array *)
        (* Check for minimul array limit value not necessary *)
        IF (diInpArrLimitMax > C_iNumOfInputCont) THEN
            wTempErrorId  := 16#1001;
            xErrorResult1 := TRUE;
        ELSIF (diOutArrLimitMax > C_iNumOfOutputCont) THEN
            wTempErrorId  := 16#1002;
            xErrorResult1 := TRUE;
        END_IF
        (* A parameter for the proof test interval is necessary. Missing this input causes an error *)
        IF (I_udiTestInterv <= 0) THEN
            wTempErrorId  := 16#1003;
            xErrorResult1 := TRUE;
        END_IF
        (* Missing the confirmation of the first proof test date causes an error *)
        IF (D#1970-1-1 = R_datTestStart) THEN
            wTempErrorId  := 16#1004;
            xErrorResult1 := TRUE;	
        END_IF
        (* Wrong input number configured causes an error *)
        FOR diLoopTrig := 1 TO diInpArrLimitMax  DO	
            (* Input 1 *)
            xInputValid := FALSE;
            CASE BYTE_TO_USINT(I_astControlInp[diLoopTrig].byMonitorInput1) OF
                12: xInputValid := TRUE;
                13:	xInputValid := TRUE;
                22: xInputValid := TRUE;
                23:	xInputValid := TRUE;
                32: xInputValid := TRUE;
                33:	xInputValid := TRUE;
                42: xInputValid := TRUE;
                43:	xInputValid := TRUE;
                52: xInputValid := TRUE;
                53:	xInputValid := TRUE;
                62: xInputValid := TRUE;
                63:	xInputValid := TRUE;
                0:	xInputValid := TRUE;
            END_CASE
            IF NOT xInputValid THEN
                wTempErrorId  := DINT_TO_WORD(16#1050 + diLoopTrig);
                xErrorResult1 := TRUE;
            END_IF
		
            (* Input 2 *)
            xInputValid := FALSE;
            CASE BYTE_TO_USINT(I_astControlInp[diLoopTrig].byMonitorInput2) OF
                12: xInputValid := TRUE;
                13:	xInputValid := TRUE;
                22: xInputValid := TRUE;
                23:	xInputValid := TRUE;
                32: xInputValid := TRUE;
                33:	xInputValid := TRUE;
                42: xInputValid := TRUE;
                43:	xInputValid := TRUE;
                52: xInputValid := TRUE;
                53:	xInputValid := TRUE;
                62: xInputValid := TRUE;
                63:	xInputValid := TRUE;
                0:	xInputValid := TRUE;
            END_CASE
            IF NOT xInputValid THEN
                wTempErrorId  := DINT_TO_WORD(16#1050 + diLoopTrig);
                xErrorResult1 := TRUE;
            END_IF
        END_FOR

        (* An parameter error stops the executing and activate the error output *)
        IF xErrorResult1 THEN
            xTempError := TRUE;
        ELSE
            (* Reset help variable for error *)
            xTempError := FALSE;
		
            (* Get active contacts and detect an reset of single counters *)
            (* Input contacts *)
            FOR diLoopIndex := 1 TO diInpArrLimitMax  DO
                (* Reset single counter (remanent variable) *)
                IF I_astControlInp[diLoopIndex].xReset THEN
                    R_audiCountInp[diLoopIndex] := 0; 
                END_IF
                (* Set BIT in array of active inputs *)
                IF (I_astControlInp[diLoopIndex].udiMaxNumOp > 0) THEN
                    axActiveInp[diLoopIndex] := TRUE;
                ELSE
                    axActiveInp[diLoopIndex] := FALSE;
                END_IF		
            END_FOR
            (* Process input *)
            (* Reset single counter (remanent variable) *)
            IF I_stControlProc.xReset THEN
                R_udiCountProc := 0;
            END_IF
            (* Set BIT of active inputs *)
            IF (I_stControlProc.udiMaxNumOp > 0) THEN
                xActiveProc := TRUE;		
            ELSE
                xActiveProc := FALSE;
            END_IF
            (* Output contacts *)
            FOR  diLoopIndex := 1 TO diOutArrLimitMax DO
                (* Reset single counter (remanent variable) *)
                IF I_astControlOut[diLoopIndex].xReset THEN
                    R_audiCountOut[diLoopIndex] := 0; 
                END_IF
                (* Set BIT in array of active inputs *)
                IF (I_astControlOut[diLoopIndex].udiMaxNumOp > 0) THEN
                    axActiveOut[diLoopIndex] := TRUE;
                ELSE
                    axActiveOut[diLoopIndex] := FALSE;
                END_IF
            END_FOR	
            (*-------------------------- Counter -----------------------------------------------------------*)	
            (* Detecting new diagnostic code input for OPERATIONAL *)	
			
            (* Be aware to use remanent variables for calculation and transfer to the output at the end *)	
            (* Increase counter depending to diagnostic code trigger *)
            IF xRTrig_Operational_Q THEN
                (* Rising edge of Operational status of the Preventa Safety Module increases all counters *) 
                (* calculation of the remaining operation counters is done separate, calculation is done until the 
                remanant counter value reaches the maximum - no LIMIT violation *)	
                (* Summarized operation counter *)
                R_udiCountSystem := R_udiCountSystem +1 ;		
                (* Process input *)
                IF xActiveProc AND (R_udiCountProc < I_stControlProc.udiMaxNumOp) THEN
                    (* Calculate result for process input and detect if maximum reached. *)
                    R_udiCountProc := R_udiCountProc + 1;				
                END_IF
                (* Array of output contacts - only counters of activated outputs will be increased*)
                FOR diLoopCountOut := 1 TO diOutArrLimitMax DO
                    IF axActiveOut[diLoopCountOut] AND (R_audiCountOut[diLoopCountOut] < I_astControlOut[diLoopCountOut].udiMaxNumOp) THEN
                        (* Counter will be increased *)
                        R_audiCountOut[diLoopCountOut] := R_audiCountOut[diLoopCountOut] + 1;					
                    END_IF
                END_FOR
            END_IF

            IF xTrig_CodeChanged AND NOT xLastCodeWas0 THEN
                (* Check which input is affected *)	
                CASE udiStatus_valid OF
                    53:	uiInputChanged := 12; (* CMD26 = input 12 = bit value 53 *)
                    52:	uiInputChanged := 13; (* CMD27 = input 13 = bit value 52 *)
                    60:	uiInputChanged := 22; (* CMD28 = input 22 = bit value 60 *)
                    61:	uiInputChanged := 23; (* CMD29 = input 23 = bit value 61 *)
                    63:	uiInputChanged := 32; (* CMD30 = input 32 = bit value 63 *)
                    62:	uiInputChanged := 33; (* CMD31 = input 33 = bit value 62 *)
                    58:	uiInputChanged := 42; (* CMD32 = input 42 = bit value 58 *)
                    59:	uiInputChanged := 43; (* CMD33 = input 43 = bit value 59 *)
                    57:	uiInputChanged := 52; (* CMD34 = input 52 = bit value 57 *)
                    56:	uiInputChanged := 53; (* CMD35 = input 53 = bit value 56 *)
                    40:	uiInputChanged := 62; (* CMD36 = input 62 = bit value 40 *)
                    41:	uiInputChanged := 63; (* CMD37 = input 63 = bit value 41 *)
                    ELSE
                    uiInputChanged := 0; (* not relevant *)
                END_CASE
		
                IF uiInputChanged <> 0 OR xCountAllInputs THEN
                    (* Array of input contacts - only counters of activated inputs will be increased *)	
                    FOR diLoopTrig := 1 TO diInpArrLimitMax DO			
                        IF axActiveInp[diLoopTrig] AND (R_audiCountInp[diLoopTrig] < I_astControlInp[diLoopTrig].udiMaxNumOp ) THEN
                            (* Counter will be increased, if this code is relevant for the input *)
                            IF BYTE_TO_USINT(I_astControlInp[diLoopTrig].byMonitorInput1) = uiInputChanged OR BYTE_TO_USINT(I_astControlInp[diLoopTrig].byMonitorInput2) = uiInputChanged OR xCountAllInputs THEN
                                R_audiCountInp[diLoopTrig] := R_audiCountInp[diLoopTrig] + 1;	
                            END_IF					
                        END_IF			
                    END_FOR
                    xCountAllInputs := FALSE;
                END_IF
            END_IF		

(* calculation for remaining counters is done for output assignment, Maximum value defined at contact 
   input is decreased BY the remanent counter value *)
   		(* Process input *)
            IF xActiveProc THEN
                IF (R_udiCountProc <= I_stControlProc.udiMaxNumOp) THEN
                    (* Substract remanent counter value of the maximum counter input value *)
                    stTempRemainNumOp.udiNumOpRemainProc  := I_stControlProc.udiMaxNumOp - R_udiCountProc;
                END_IF
                (* Remaining counter result of ZERO occurs an error *)
                IF (0 = stTempRemainNumOp.udiNumOpRemainProc) THEN
                    xOpExceeded := TRUE;
                    wTempExceededId  := 16#1005 ;		
                END_IF				
            END_IF
		
            (* Array of active input contacts *)
            FOR diLoopIndex := 1 TO diInpArrLimitMax DO			
                IF axActiveInp[diLoopIndex] THEN
                    IF (R_audiCountInp[diLoopIndex] <= I_astControlInp[diLoopIndex].udiMaxNumOp ) THEN
                        (* Substract remanent counter value of the maximum counter input value *)
                        stTempRemainNumOp.audiNumOpRemainInp[diLoopIndex] := I_astControlInp[diLoopIndex].udiMaxNumOp - R_audiCountInp[diLoopIndex];
                    END_IF
                    (* Remaining counter result of ZERO occurs an error, each contact has own error ID *)
                    IF (0 = stTempRemainNumOp.audiNumOpRemainInp[diLoopIndex]) THEN
                        xOpExceeded := TRUE;
                        wTempExceededId  := DINT_TO_WORD(16#1010 + diLoopIndex) ;
                    END_IF
                END_IF	
            END_FOR
		
            (* Array of output contacts *)
            FOR diLoopIndex := 1 TO diOutArrLimitMax  DO			
                IF axActiveOut[diLoopIndex] THEN
                    IF (R_audiCountOut[diLoopIndex] <= I_astControlOut[diLoopIndex].udiMaxNumOp) THEN
                        (* Substract remanent counter value of the maximum counter input value *)
                        stTempRemainNumOp.audiNumOpRemainOut[diLoopIndex] := I_astControlOut[diLoopIndex].udiMaxNumOp - R_audiCountOut[diLoopIndex];
                    END_IF
                    (* Remaining counter result of ZERO occurs an error, each contact has own error ID *)
                    IF (0 = stTempRemainNumOp.audiNumOpRemainOut[diLoopIndex]) THEN
                        xOpExceeded := TRUE;
                        wTempExceededId  := DINT_TO_WORD(16#1020 + diLoopIndex) ;	
                    END_IF		
                END_IF
            END_FOR
		
            (* An error occurs, help variable for status output is activated *)
            IF xErrorResult2 THEN
                xTempError := TRUE;
            ELSE
                xTempError := FALSE;
            END_IF
            (*---------------------------- Proof test -----------------------------------------------------------*)				
            (* Restore an input array of proof test timestamps to the internal array *)
            RTrig_Restore(CLK:= I_xSaveBProofTests); 
            xRTrig_Restore_Q := RTrig_Restore.Q;
            IF xRTrig_Restore_Q THEN
                R_adtProof := I_adtProofTests;
            END_IF
		
            (* Store executed proof test timestamps in an remanent array *)
            RTrig_TestDone(CLK:=I_xTestExecuted);
            IF RTrig_TestDone.Q THEN
                // Move values in arrays and restore new TOD
                FOR iLoop := C_iProofTestStorage TO 1 BY -1 DO
                    IF (iLoop = 1) THEN						
                        R_adtProof[1]  	:= dtTestIntervalRef;
                    ELSE
                        R_adtProof[iLoop]  	:= R_adtProof[iLoop-1];
                    END_IF
                END_FOR			
            END_IF
            (* Calculate next proof date with parameter I_udiTestInterv *)
            udiCalcDaysSec := I_udiTestInterv *24*60*60; // Days -> hours -> minutes -> seconds
            udiCalcNewDate := DATE_TO_UDINT(DT_TO_DATE(R_adtProof[1])) + udiCalcDaysSec;
            datNextProofTest:= UDINT_TO_DATE(udiCalcNewDate); 
            (* Check if next proof date is violated, because current date is more than the result *)	
            IF (DT_TO_DATE(dtTestIntervalRef) > datNextProofTest) THEN
                xProofDateViolated := TRUE;
            ELSE
                xProofDateViolated := FALSE;
            END_IF		
        END_IF
    END_IF
	
    IF NOT xTempError THEN
        wTempErrorId := 0;
    END_IF

    (* Define outputs *)
    // Function block status outputs
    Q_xError 	:= xTempError;
    Q_wErrorId	:= wTempErrorId;
    Q_xActive 	:= I_xEnable;
    Q_xOpExceeded := xOpExceeded;
    Q_wExceededId := wTempExceededId;
    // Remaining operation values
    Q_udiNumOpSystem	:= R_udiCountSystem;
    Q_stRemainNumOp		:= stTempRemainNumOp;		
    //Proof test outputs
    Q_datTestStart		:= R_datTestStart;
    Q_datNextProof 		:= datNextProofTest;
    Q_xTestIntervViol	:= xProofDateViolated;
    Q_adtProofTests 	:= R_adtProof;	
	
END_FUNCTION_BLOCK