
FUNCTION_BLOCK FB_CheckCallInInterval (* *) (*$GROUP=User,$CAT=User,$GROUPICON=User.png,$CATICON=User.png*)
	VAR_INPUT
		I_xEnable : BOOL; (*Enables the function block, check of a cyclic call being executed.*)
		I_timCycleTime : TIME; (*Parameter for cycle time of the superordinated function/ function block.*)
	END_VAR
	VAR_OUTPUT
		Q_xCallFailed : BOOL; (*Error information. Set TRUE if the cyclic call failed.*)
	END_VAR
	VAR
		Ton_CycleTime : TON; (*Timer started with parameter Task cycle time to check of valid cyclic call.*)
		iCycleCnt : INT; (*Help counter for next cycle.*)
		xCallError : {REDUND_UNREPLICABLE} BOOL; (*Help variable to detect intern errors and activate output*)
	END_VAR
END_FUNCTION_BLOCK

FUNCTION_BLOCK FB_CheckCode (* *) (*$GROUP=User,$CAT=User,$GROUPICON=User.png,$CATICON=User.png*)
	VAR_INPUT
		I_iCode : INT; (*code to be ckecked*)
	END_VAR
	VAR_OUTPUT
		Q_xValid : BOOL; (*Is the code valid?*)
		Q_dwCode1_32 : DWORD; (*DWORD with codes 1..32*)
		Q_dwCode33_64 : DWORD; (*DWORD with codes 33..64*)
	END_VAR
	VAR CONSTANT
		C_iCodeCmd1 : INT := 0; (*value of bytesequence of CMD 1 - Module status code*)
		C_iCodeCmd2 : INT := 45; (* value of bytesequence of CMD 2 - Module status code *)
		C_iCodeCmd3 : INT := 3; (* value of bytesequence of CMD 3- Module status code *)
		C_iCodeCmd4 : INT := 6; (* value of bytesequence of CMD 4 - Module status code *)
		C_iCodeCmd5 : INT := 7; (* value of bytesequence of CMD 5- Module status code *)
		C_iCodeCmd6 : INT := 12; (* value of bytesequence of CMD 6 - Module status code *)
		C_iCodeCmd7 : INT := 13; (* value of bytesequence of CMD 7 - Module status code *)
		C_iCodeCmd8 : INT := 15; (* value of bytesequence of CMD 8 - Module status code *)
		C_iCodeCmd9 : INT := 14; (* value of bytesequence of CMD 9 - Module status code *)
		C_iCodeCmd10 : INT := 24; (* value of bytesequence of CMD 10 - Module status code *)
		C_iCodeCmd11 : INT := 44; (* value of bytesequence of CMD 11 - Module status code *)
		C_iCodeCmd12 : INT := 27; (* value of bytesequence of CMD 12 - Module status code *)
		C_iCodeCmd13 : INT := 26; (* value of bytesequence of CMD 13 - Module status code *)
		C_iCodeCmd14 : INT := 30; (* value of bytesequence of CMD 14 - Module status code *)
		C_iCodeCmd15 : INT := 31; (* value of bytesequence of CMD 15 - Module status code *)
		C_iCodeCmd16 : INT := 29; (* value of bytesequence of CMD 16 - Module status code *)
		C_iCodeCmd17 : INT := 28; (* value of bytesequence of CMD 17 - Module status code *)
		C_iCodeCmd18 : INT := 48; (* value of bytesequence of CMD 18 - Module status code *)
		C_iCodeCmd19 : INT := 35; (* value of bytesequence of CMD 19 - Module status code *)
		C_iCodeCmd20 : INT := 38; (* value of bytesequence of CMD 20 - Module status code *)
		C_iCodeCmd21 : INT := 32; (* value of bytesequence of CMD 21 - Module status code *)
		C_iCodeCmd22 : INT := 51; (* value of bytesequence of CMD 22 - Module status code *)
		C_iCodeCmd23 : INT := 39; (* value of bytesequence of CMD 23 - Module status code *)
		C_iCodeCmd24 : INT := 54; (* value of bytesequence of CMD 24 - Module status code *)
		C_iCodeCmd25 : INT := 55; (* value of bytesequence of CMD 25 - Module status code *)
		C_iCodeCmd26 : INT := 53; (* value of bytesequence of CMD 26 - Module status code *)
		C_iCodeCmd27 : INT := 52; (* value of bytesequence of CMD 27 - Module status code *)
		C_iCodeCmd28 : INT := 60; (* value of bytesequence of CMD 28 - Module status code *)
		C_iCodeCmd29 : INT := 61; (* value of bytesequence of CMD 29 - Module status code *)
		C_iCodeCmd30 : INT := 63; (* value of bytesequence of CMD 30 - Module status code *)
		C_iCodeCmd31 : INT := 62; (* value of bytesequence of CMD 31 - Module status code *)
		C_iCodeCmd32 : INT := 58; (* value of bytesequence of CMD 32 - Module status code *)
		C_iCodeCmd33 : INT := 59; (* value of bytesequence of CMD 33 - Module status code *)
		C_iCodeCmd34 : INT := 57; (* value of bytesequence of CMD 34 - Module status code *)
		C_iCodeCmd35 : INT := 56; (* value of bytesequence of CMD 35 - Module status code *)
		C_iCodeCmd36 : INT := 40; (* value of bytesequence of CMD 36 - Module status code *)
		C_iCodeCmd37 : INT := 41; (* value of bytesequence of CMD 37 - Module status code *)
		C_iCodeCmd38 : INT := 43; (* value of bytesequence of CMD 38 - Module status code *)
		C_iCodeCmd39 : INT := 42; (* value of bytesequence of CMD 39 - Module status code *)
		C_iCodeCmd40 : INT := 46; (* value of bytesequence of CMD 40 - Module status code *)
		C_iCodeCmd41 : INT := 47; (* value of bytesequence of CMD 41 - Module status code *)
	END_VAR
END_FUNCTION_BLOCK

FUNCTION_BLOCK FB_PreventaDiag (* *) (*$GROUP=User,$CAT=User,$GROUPICON=User.png,$CATICON=User.png*)
	VAR_INPUT
		I_xEnable : BOOL; (* TRUE activates the function block. While activated, the bit sequence at input i_xDiagSignal is continuously read and evaluated. *)
		I_xDiagSignal : BOOL; (* Input for connecting the bit sequence output by the Preventa XPS safety module. Connect this input to the variable which is assigned to the auxiliary output Z1 signal. *)
		I_timTaskCycle : TIME; (* Currently configured cycle time in ms of the cyclic task in which the function block is executed. *)
	END_VAR
	VAR_OUTPUT
		Q_xBusy : BOOL; (* If this output is set to TRUE, the function block execution is in progress. *)
		Q_xValid : BOOL; (* The output value TRUE indicates for one cycle a valid bit sequence at the i_xDiagSignal input. *)
		Q_xError : BOOL; (* If this output is set to TRUE, the function block has detected an error. For details, refer to q_wErrorId. *)
		Q_stDiagCode : ST_DiagCodes; (* Diagnostic information decoded from the input bit sequence: The corresponding bit value in the ST_DiagCodes structure are set to TRUE. *)
		Q_dwStatus : DWORD; (* Indicates the 6-bit code part (without the 4-bit start part) of the last detected bit sequence converted to a double word data type. *)
		Q_xComWireInShort : BOOL; (* The output value TRUE indicates a static TRUE signal at input i_xDiagSignal. *)
		Q_xComWireOpen : BOOL; (* The output value TRUE indicates a static FALSE signal input i_xDiagSignal. *)
		Q_wErrorId : WORD; (* Provides diagnostic and status information as a numeric value. *)
	END_VAR
	VAR CONSTANT
		C_xDelayOneFrame : BOOL := FALSE; (*delay the output of a received code by one frame. If the next code is 0, do nothing*)
		C_iStateIdle : INT := 0; (* State machine - IDLE, waiting for start *)
		C_iStateInit : INT := 10; (* State machine - INIT, parameter check *)
		C_iStateSearch : INT := 20; (* State machine - Search, parameter check *)
		C_iStateCode : INT := 30; (* State machine - Code, starting analyzing binary input and detecting a code sequence *)
		C_iStateError : INT := 100; (* State machine - Error, error occurs, send status output *) (* Constants for START and CODE sequence properties	 *)
		C_siNumBitStartSq0 : SINT := 4; (* Number of bits for start sequence *)
		C_siNumBitCodeSeq : SINT := 6; (* Number of bits used for one diagnostic code *)
		C_timTaskCycleTime : TIME := T#50MS; (* Task cycle time must be set for detecting the signal change in bit duraiton cycle *)
		C_timSeqTotalMax : TIME := T#2200MS; (* Maximum time to wait for the frame to be complete *)
		C_timTimeOutStart : TIME := T#7S; (* Timeout to control an detection of a start sequence *)
		C_iBitDuration : INT := 200; (* Bit duraiton of 200ms for one new sequence bit *)
		C_iMaxInvSearchFrames : INT := 4; (* Number of invalid frames during frame search, before an error is triggered *)
		C_iMaxInvCodeFrames : INT := 1; (* Number of invalid frames during normal operation, before an error is triggered *) (* Constants for Error numbers *)
		C_wERROR_none : WORD := 16#0000; (* No error *)
		C_wERROR_MissedCall : WORD := 16#1000; (* Function block is not called cyclically *)
		C_wERROR_CycleTime : WORD := 16#1001; (* Parametrization error of task cycle time; must be lower or equal than the defined constant C_timTaskCycleTime *)
		C_wERROR_SearchTOut : WORD := 16#1002; (* Timeout for detecting a START sequence; Timeout defined as constant C_timTimeOutStart *)
		C_wERROR_TooManyBits : WORD := 16#1003; (* Number OF detected bits in CODE sequence is higher than defined value in C_siNumBitCodeSeq *)
		C_wERROR_InvalidCode : WORD := 16#1004; (* Detected diagnostic code is not defined *)
		C_wERROR_InvalidState : WORD := 16#1005; (* Error in defined state machine value *)
	END_VAR
	VAR
		iState : INT; (*State machine status*)
		timCycleTimeLastInit : TIME; (* Help variable to compare the current parametrized cycle time with the time of the last time, the cycle time was checked for validity  *)
		xFrameBuffer : ARRAY[1..10] OF BOOL; (* Holding the frame sent by the preventa module *)
		iActualBit : INT; (* Number of the active bit on the signal line *)
		iPreviousBit : INT; (* Number of the previous bit on the signal line *)
		iActInvalidFrames : INT; (* Counter for invalid frames during frame search *)
		TON_TickerS : TON; (* Time ticker for start sequence, time defined as constant value *)
		xTON_TickerS_Q : BOOL; (* Result of time ticker for start sequence, TRUE = time elapsed *)
		timTON_TickerS_ET : TIME; (* Current time elapsed since timer started *)
		TON_TimeoutS : TON; (* Timeout timer, if no start sequence can be detect in the defined time *)
		xTon_TimeoutS_Q : BOOL; (* Result of the timeout to activate an error *)
		R_Trigger : R_TRIG;
		fbCheckCall : FB_CheckCallInInterval;
		xErrorFbCall : BOOL;
		fbCheckCode : FB_CheckCode; (* function to check if a code is existing *)
		wFrame : WORD; (* Complete frame packed in one word *)
		byCodeSequence : BYTE; (* Code sequence describes diagnostic code *)
		byCodeSeqOut : BYTE; (* Codesequence to send as output, only rewrite if code could be confirmed *)
		iCode : INT; (* CASE_OF variable for Code sequence validation *)
		xInternError : BOOL; (* Help status variable for ERROR *)
		xInternBusy : BOOL; (* Help status variable for BUSY *)
		xInternValid : BOOL; (* Help status variable  for VALID *)
		xShortCircuitOn : BOOL; (* help variable to detect short circuit for defining status output *)
		xPowerOff : BOOL; (* help variable to detect power supply is off *)
		wErrorId : WORD; (* Temporary value of the error code as detailed description to the error output status *)
		dwCode1_32_tmp : DWORD; (* Temporary storage for 31 diagnostic codes in 32bit *)
		dwCode33_64_tmp : DWORD; (* Temporary storage for 31 diagnostic codes in 32bit *)
		dwCode1_32 : DWORD; (* Storage for 31 diagnostic codes in 32bit *)
		dwCode33_64 : DWORD; (* Storage for 31 diagnostic codes in 32bit *)
		iEdgeOffset : INT; (* Offset of detected edge to the desired smapling position*)
		iEdgeOffset2 : INT; (* slowly incremented / decremented offset, to avout big jumps *)
		xEdgeDetected : BOOL; (* TRUE if a rising edge was detected on the input signal*)
		xCodeValid : BOOL; (* TRUE if the code captured during search is valid / existing *)
		xCaptureSignal : BOOL; (* Trigger signal to capture a single bit *)
		xNewFrame : BOOL; (* New frame captured *)
		xLockSearch : BOOL; (* Stay in search mode until at least 10 bits have been received *)
		dwCode1_32_Buffered : DWORD; (* Storage for 31 diagnostic codes in 32bit *)
		dwCode33_64_Buffered : DWORD; (* Storage for 31 diagnostic codes in 32bit *)
		byCode_Buffered : BYTE; (* Code sequence describes diagnostic code *)
		xSendValidOnce : BOOL; (* Send Valid signal one time after power loss was detected *)
		xErrorCode : BOOL; (* Temporary error result in 'Code sequence' for no Power supply detected *)
		xNewStart : BOOL; (*previous state was xEnable = FALSE*)
		xStartSeqFound : BOOL; (*Block a rising edge on the signal from being detectet*)
	END_VAR
END_FUNCTION_BLOCK

FUNCTION_BLOCK FB_PreventaMain (* *) (*$GROUP=User,$CAT=User,$GROUPICON=User.png,$CATICON=User.png*)
	VAR_INPUT
		I_xEnable : BOOL; (* TRUE activates the function block. While activated, the diagnostic code at input i_dwStatus is read and evaluated. *)
		I_dwStatus : DWORD; (* Connect this input directly to the q_dwStatus output of the FB_PreventaDiag function block. *)
		I_xValid : BOOL; (* Connect this input directly to the q_xValid output of the FB_PreventaDiag function block. *)
		I_stControlProc : ST_DevControl; (* Defines the threshold value for operating cycles of the safety contacts channels in the Preventa XPS safety module and specifies the devices connected to the module outputs. *)
		I_astControlInp : ARRAY[1..6] OF ST_InputControl; (* Specification of the input devices connected to the Preventa XPS safety module. *)
		I_astControlOut : ARRAY[1..13] OF ST_DevControl; (* Specification of the output devices connected to the Preventa XPS safety module. *)
		I_dtTestIntervalRef : DATE_AND_TIME; (* Date and time of the proof test to be executed. *)
		I_udiTestInterv : UDINT; (* Proof test interval. Enter the number of days between two proof tests. *)
		I_xNewModuleActive : BOOL; (* Edge-triggered input to confirm the execution of the first proof test for a newly added or replaced module. *)
		I_xTestExecuted : BOOL; (* Edge-triggered input to confirm the executed proof test. *)
		I_xSaveBProofTests : BOOL; (* Edge-triggered input to read the time stamps from the proof test array connected to the input i_adtProofTests. *)
		I_adtProofTests : ARRAY[1..10] OF DATE_AND_TIME; (* External proof test array containing proof test time stamps which have been, for example, generated in and provided by an HMI application. *)
	END_VAR
	VAR_OUTPUT
		Q_xActive : BOOL; (* If this output is set to TRUE, the function block execution is in progress. *)
		Q_xError : BOOL; (* If this output is set to TRUE, the fucntion block has detected an error. For details, refer to q_wErrorId. *)
		Q_datTestStart : DATE; (* Confirmed date value of the first executed proof test. *)
		Q_datNextProof : DATE; (* Calculated date for the execution of the next proof test. *)
		Q_xTestIntervViol : BOOL; (* The output value TRUE indicates that the proof test interval has benn exceeded. *)
		Q_xOpExceeded : BOOL; (* The output value TRUE indicates that the number of remaining operating cycles is 0 for at least one monitored device or safety contacts channel of the Preventa XPS safety module. *)
		Q_wExceededId : WORD; (* ID of the input/output that exceeded its operation. *)
		Q_adtProofTests : ARRAY[1..10] OF DATE_AND_TIME; (* Array of time stamps of the last ten executed proof tests. *)
		Q_udiNumOpSystem : UDINT; (* Number of operating cycles of the entire safety system, including the safety contacts channels in the Preventa XPS safety module. *)
		Q_stRemainNumOp : ST_RemainNumOp; (* Indicates the result of the maintenance calculation. *)
		Q_wErrorId : WORD; (* Provides diagnostic and status information as a numeric value. *)
	END_VAR
	VAR CONSTANT
		C_iNumOfInputCont : INT := 6; (* Number for maximum of input contacts *)
		C_iNumOfOutputCont : INT := 13; (* Number of maximum of output contacts *)
		C_iProofTestStorage : INT := 10; (* Array size for storage of timestamps of executed proof test  *)
		C_udiCMD_Operational : UDINT := 47; (* CMD41 = operational = bit value 47 *)
		C_udiCMD_AllInpOff : UDINT := 55; (*Command all inputs off*)
	END_VAR
	VAR RETAIN
		R_udiCountSystem : UDINT; (* Increasing counter for system operations *)
		R_audiCountInp : ARRAY[1..C_iNumOfInputCont] OF UDINT; (* Increasing of input contact counter (Stored in an array), coutner are limited by the udiMaxNumOp parameter *)
		R_udiCountProc : UDINT; (* Increasing counter of process unit, limited by the udiMaxNumOp parameter *)
		R_audiCountOut : ARRAY[1..C_iNumOfOutputCont] OF UDINT; (* Increasing of output contact counter (Stored in an array), coutner are limited by the udiMaxNumOp parameter *)
		R_adtProof : ARRAY[1..C_iProofTestStorage] OF DATE_AND_TIME; (* Array of 10 timestamps of executed proof tests *)
		R_datTestStart : DATE; (* Timestamp of the first test the module is commissioned *)
	END_VAR
	VAR
		xRTrig_Operational_Q : BOOL; (* Status output, trigger for rising edge detected *)
		axActiveInp : ARRAY[1..C_iNumOfInputCont] OF BOOL; (* Array to save check results of active inputs *)
		xActiveProc : BOOL; (* Check result of the active process contact *)
		axActiveOut : ARRAY[1..C_iNumOfOutputCont] OF BOOL; (* Array to save check results of active outputs *)
		wTempErrorId : WORD; (* Local Error ID for transfer to the output *) (* Help trigger for detecting a new module with the first proof test timestamp *)
		RTrig_ModuleExch : R_TRIG; (* Trigger of rising edge for a new module - saving new proof test start timestamp and reset remanent counters *)
		xRTrig_ModuleExch_Q : BOOL; (* Status ouput of the trigger *) (* calculation remaining values *)
		stTempRemainNumOp : ST_RemainNumOp; (* Structure of calculation results for the remaining number of operations  to transfer to the output *)
		diInpArrLimitMin : DINT; (* Value of the minimum limit for input contact array *)
		diInpArrLimitMax : DINT; (* Value of the maximum limit for input contact array *)
		diOutArrLimitMin : DINT; (* Value of the minimum limit for output contact array *)
		diOutArrLimitMax : DINT; (* Value of the maximum limit for output contact array *) (* help variables for proof test interval *)
		datNextProofTest : DATE; (* Date value of the necessary next proof test execution *)
		udiCalcNewDate : UDINT; (* New proof test date calculated *)
		udiCalcDaysSec : UDINT; (* Calculated value of hours in seconds (DWORD time resolution) for calculation in DWORD and convertion to DATE *)
		xProofDateViolated : BOOL; (* Help variable set TRUE, if current timestamp > value of next proof date *)
		RTrig_TestDone : R_TRIG; (* Trigger of rising edge to save a new timestamp into array of proof test *)
		RTrig_Restore : R_TRIG; (* Trigger of rising edge to save back an input array of proof test timestamps to the internal array. *)
		xRTrig_Restore_Q : BOOL; (* Status output of the trigger. *) (* Loop counter index *)
		diLoopIndex : DINT; (* Help variable for loop index as DINT *)
		diLoopCountOut : DINT; (* Help variable for counting output contacts while operational state has detected *)
		iLoopReset : INT; (* Help variable for reset remanent counters while a new module is activated *)
		diLoopTrig : DINT; (* Help varibale for counting input concact operations depening on trigger output status *)
		iLoop : INT; (* Loop index to fill up proof test date array *)
		udiStatus_valid : UDINT; (* Actual valid status received *)
		udiStatus_old : UDINT; (* Last valid status received *)
		xTrig_CodeChanged : BOOL; (* Code did change, chack if any input affected *)
		uiInputChanged : USINT; (* Input that changed it's level *)
		xOpExceeded : BOOL; (* Number of operations for any signal exceeded *)
		wTempExceededId : WORD; (* ID of the input/output that exceeded it's operation count *)
		xInputValid : BOOL; (* Input is existing/valid *)
		xErrorResult1 : BOOL; (* Temporary error result for parameter check *)
		xErrorResult2 : BOOL; (* Temporary error result *)
		xTempError : BOOL; (* Local status of error for transfer to the output *)
		xCountAllInputs : BOOL; (* Count all inputs if change from CMD47 to CMD25 *)
		xLastCodeWas0 : BOOL; (* TRUE if the code changed from 0 to x *)
		DTGetTime_0 : DTGetTime;
		dtTestIntervalRef : DATE_AND_TIME;
	END_VAR
END_FUNCTION_BLOCK
